"""A module for simple checks to be used within code and testing.

The easycheck module offers simple functions to check conditions and either raise
an exception or issue a warning if the condition is violated; otherwise,
nothing happens (the function returns None). You can either choose default
exceptions and messages (or no message) or customize them. Unlike the assert
statement, you can use easycheck functions within code.

The module also offers aliases to be used in testing, all of which have the
word "assert" in their names (assert_if(), assert_if_not(),
assert_type(), assert_length(), and assert_path()).
"""
from __future__ import annotations

import builtins
import os
import warnings

from collections import abc
from functools import wraps
from math import isclose
from numbers import Number
from operator import eq, le, lt, gt, ge, ne, is_, is_not
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union


# Type alias and variables

PathType = Union[Path, str]
T = TypeVar("T")


# Exceptions


class LimitError(Exception):
    """Number out of limit."""


class LengthError(Exception):
    """Violated length check."""


class ComparisonError(Exception):
    """The comparison is not true."""


class NotCloseEnoughError(Exception):
    """The two float numbers are not close enough."""


class ArgumentValueError(Exception):
    """Argument's value is incorrect."""


def switch(func: abc.Callable) -> abc.Callable:
    """Decorator to switch off all easycheck checks.

    It does so by getting the EASYCHECK_RUN environmental variable.
    When it's set to "0", easycheck is switched off.
    """

    @wraps(func)
    def inner(*args: Any, **kwargs: Any) -> Any:
        if os.environ.get("EASYCHECK_RUN", 1) != "0":
            return func(*args, **kwargs)

    return inner


@switch
def check_if(
    condition: bool, handle_with: type = AssertionError, message: Optional[str] = None
) -> None:
    """Check if a condition is true.

    Args:
        condition (bool): condition to check.
        handle_with (type): the type of exception to be raised or warning to
            be issued
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter,
        AssertionError by default (unless handle_with is a warning).

    >>> check_if(2 > 1)
    >>> check_if(2 < 1)
    Traceback (most recent call last):
        ...
    AssertionError
    >>> check_if(2 < 1,
    ...    handle_with=ValueError,
    ...    message='2 is not smaller than 1')
    Traceback (most recent call last):
        ...
    ValueError: 2 is not smaller than 1

    with its shorter version of
    >>> check_if(2 < 1, ValueError, '2 is not smaller than 1')
    Traceback (most recent call last):
        ...
    ValueError: 2 is not smaller than 1

    You can also make the condition a complex condition:
    >>> args = 33, 275, 'fifty-four'
    >>> check_if(args[0] < 50 and args[1] > 100 and isinstance(args[2], str))

    This might be more readable using unpacking (particularly if each element
    of the tuple has meaning):
    >>> a, b, c = args
    >>> check_if(a < 50 and b > 100 and isinstance(c, str))

    or
    >>> check_args = a < 50 and b > 100 and isinstance(c, str)
    >>> check_if(check_args)

    In such combined comparisons, you can easily use any logical operator:
    >>> check_if((a < 50 and b > 100) or isinstance(c, str))

    To issue a warning, use the Warning class or its subclass
    (we'll catch the warning here):
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_if(2 < 1, handle_with=Warning, message='2 is not smaller than 1')
    ...     assert_if("2 is not smaller than 1" in str(w[-1].message))

    or shorter
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_if(2 < 1, Warning, '2 is not smaller than 1')
    ...     assert_if("2 is not smaller than 1" in str(w[-1].message))
    """
    __tracebackhide__ = True
    if not condition:
        _raise(handle_with, message)


@switch
def check_if_not(
    condition: bool, handle_with: type = AssertionError, message: Optional[str] = None
) -> None:
    """Check if a condition is not true.

    Args:
        condition (bool): condition to check.
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.


    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter,
        AssertionError by default (unless handle_with is a warning).

    You would normally use these functions in situations like these: This is
    engine speed in the object engine_speed:
    >>> engine_speed = 5900

    In this case, assume that if this value is higher than 6000, than this
    situation may cause problems. So, let us check this:

    >>> check_if_not(engine_speed > 6000, ValueError, 'Danger!')

    Sure, you can do so using the check_if() function, like here:
    >>> check_if(engine_speed <= 6000, ValueError, 'Danger!')

    and both are fine. You simply have two functions to choose from in order to
    make the code as readable as you want. Sometimes the negative version
    sounds more natural, and so it's all about what kind of language you want
    to use in this particular situation.

    Consider the examples below:
    >>> check_if_not(2 == 1)
    >>> check_if_not(2 > 1)
    Traceback (most recent call last):
        ...
    AssertionError
    >>> check_if_not(2 > 1, message="")
    Traceback (most recent call last):
        ...
    AssertionError
    >>> check_if_not(2 > 1, ValueError, '2 is not smaller than 1')
    Traceback (most recent call last):
        ...
    ValueError: 2 is not smaller than 1

    >>> BMI = 50
    >>> disaster = True if BMI > 30 else False
    >>> check_if_not(disaster, message='BMI disaster! Watch out for candies!')
    Traceback (most recent call last):
        ...
    AssertionError: BMI disaster! Watch out for candies!

    To issue a warning, use the Warning class or one of its subclasses
    (we'll catch the warning):
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_if_not(2 > 1, Warning, '2 is bigger than 1')
    ...     assert_if("2 is bigger than 1" in str(w[-1].message))
    """
    __tracebackhide__ = True
    if condition:
        _raise(handle_with, message)


@switch
def check_if_in_limits(
    x: float,
    lower_limit: float = float("-inf"),
    upper_limit: float = float("inf"),
    handle_with: type = LimitError,
    message: Optional[str] = None,
    include_equal: bool = True,
) -> None:
    """Check if number is in range of limits

    Args:
        x (float): number to be checked if it's within specified limits
        lower_limit (float): the lower limit of the interval
        upper_limit (float): the upper limit of the interval
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.
        include_equal (bool): True for strict checks (lower ≤ x ≤ upper),
        False otherwise (lower < x < upper)

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter, LengthError
        by default (unless handle_with is a warning).

    >>> check_if_in_limits(3, 1, 5)
    >>> check_if_in_limits(3, 1)
    >>> check_if_in_limits(-3, -5, -1)
    >>> check_if_in_limits(1.0, -1, 3)
    >>> check_if_in_limits(-1.0, -100.0, 100.0)
    >>> check_if_in_limits(0.0001, 0.00005, 0.0003)
    >>> check_if_in_limits(3, upper_limit = 5)
    >>> check_if_in_limits(3, 1, 3, include_equal = True)
    >>> check_if_in_limits(3, 1, 3, include_equal = False)
    Traceback (most recent call last):
        ...
    LimitError: Number out of limit.
    >>> check_if_in_limits(5, 1, 3)
    Traceback (most recent call last):
        ...
    LimitError: Number out of limit.
    >>> check_if_in_limits(5, 1, 3, Warning)
    """
    __tracebackhide__ = True
    x = float(x)
    if include_equal:
        condition = lower_limit <= x <= upper_limit
    else:
        condition = lower_limit < x < upper_limit

    if not condition:
        _raise(handle_with, message)


@switch
def check_length(
    item: Union[abc.Sized | Number],
    expected_length: int,
    handle_with: type = LengthError,
    message: Optional[str] = None,
    operator: abc.Callable = eq,
    assign_length_to_others: bool = False,
) -> None:
    """Compare item's length with expected_length, using operator.

    Args:
        item: the object whose length we want to validate
        expected_length (int): the expected length of the item
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.
        operator: one of the functions returned by get_possible_operators()
        assign_length_to_others (bool): treat all numeric types as having the
            length of 1. If False, any attempt to validate a numeric type
            will raise an exception/warning, as numeric types don't implement
            __len__()

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter, LengthError
        by default (unless handle_with is a warning).

    >>> check_length(['string'], 1)
    >>> check_length('string', 6)
    >>> check_length([1, 2], 2)
    >>> check_length(len(i for i in range(3)))
    Traceback (most recent call last):
        ...
    TypeError: object of type 'generator' has no len()
    >>> check_length(2, 1, assign_length_to_others=True)
    >>> check_length(2, 0, operator=gt, assign_length_to_others=True)
    >>> check_length(True, 1, assign_length_to_others=True)

    To issue a warning, use the Warning class or its subclass:
    >>> check_length('string', 6, Warning)
    """
    __tracebackhide__ = True
    if assign_length_to_others:
        if isinstance(item, (Number, bool)):
            item = [item]

    condition = operator(len(item), expected_length)  # type: ignore
    if not condition:
        _raise(handle_with, message)


@switch
def check_type(
    item: Any,
    expected_type: Union[type, abc.Sequence[type]],
    handle_with: type = TypeError,
    message: Optional[str] = None,
) -> None:
    """Check if item has the type of expected_type.

    Args:
        item: the object whose type we want to validate
        expected_type (type, Iterable[type]): the expected type of the item
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter, TypeError
        by default.

    If you want to check if the item is None, you can do so in two ways:
    >>> my_none_object = None
    >>> check_if(my_none_object is None, handle_with=TypeError)

    or
    >>> check_type(my_none_object, None)

    None is not a type, but it gets special treatment so that you can use
    the check_type() function for None objects.

    >>> check_type(['string'], list)
    >>> check_type('string', str)
    >>> check_type((1, 2), tuple)
    >>> check_type([1, 2], (tuple, list), message='Neither tuple nor list')
    >>> check_type('souvenir',
    ...    (tuple, list),
    ...    message='Neither tuple nor list')
    Traceback (most recent call last):
        ...
    TypeError: Neither tuple nor list
    >>> check_type((i for i in range(3)), tuple)
    Traceback (most recent call last):
        ...
    TypeError
    >>> check_type(
    ...    (i for i in range(3)), tuple, message='This is not tuple.')
    Traceback (most recent call last):
        ...
    TypeError: This is not tuple.
    >>> from collections.abc import Generator
    >>> check_type((i for i in range(3)), Generator)

    You can include None:
    >>> check_type('a', (str, None))
    >>> check_type(None, expected_type=(str, None))

    To issue a warning, do the following:
    >>> check_type('a', (str, None), Warning, 'Undesired instance')
    """
    __tracebackhide__ = True
    if expected_type is None:
        if item is not None:
            _raise(handle_with, message)
        return None

    if isinstance(expected_type, abc.Iterable):
        if item is None and any(t is None for t in expected_type):
            return None
        expected_type = tuple(t for t in expected_type if t is not None)

    if not isinstance(item, expected_type):
        _raise(handle_with, message)


@switch
def check_if_isclose(
    x: float,
    y: float,
    /,
    handle_with: type = NotCloseEnoughError,
    message: Optional[str] = None,
    rel_tol: float = 1e-09,
    abs_tol: float = 0.0,
) -> None:
    """Check if two floats are close in value.

    The function is just a wrapper around math.isclose(), and its defaults
    are exactly the same. Two values (x and y, both being positional-only
    parameters) will be considered close when the difference between them
    (either relative or absolute) is smaller than at least one of the
    tolerances. If you do not want to use any of the two tolerances, set it
    to 0.

    Note: Before applying math.isclose(), x and y are first converted to
    floats, so you can provide them as integers or even strings.

    At least one tolerance needs to be provided (so not be zero); otherwise
    the function will do nothing.

    Unlike most easycheck functions, check_if_isclose() uses two
    positional-only and four keyword-only arguments. So when providing one of
    the two tolerances, you have to specify it using the argument's name. You
    have to do the same also for handle_with and message.

    Args:
        x, y (float): two numbers to compare
        rel_tol (float): maximum difference for being considered "close",
            relative to the magnitude of the input values
        abs_tol (float): maximum difference for being considered "close",
            regardless of the magnitude of the input values
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter,
        NotCloseEnoughError by default.

    >>> check_if_isclose(1.12, 1.12, rel_tol=1e-09)
    >>> check_if_isclose('1.12', '1.12', rel_tol=1e-09)
    >>> check_if_isclose(1.12, 1.13, rel_tol=1e-09)
    Traceback (most recent call last):
        ...
    NotCloseEnoughError: The two float numbers are not close enough.

    >>> check_if_isclose(1.12, 1.13, rel_tol=.05)
    >>> check_if_isclose(1.12, 1.13, abs_tol=.05)
    >>> check_if_isclose(1.12, 1.13, abs_tol=.005)
    Traceback (most recent call last):
        ...
    NotCloseEnoughError: The two float numbers are not close enough.

    >>> check_if_isclose(1.12, 2.12)
    Traceback (most recent call last):
        ...
    NotCloseEnoughError: The two float numbers are not close enough.

    >>> check_if_isclose(1.12, 2.12, ValueError, rel_tol=1e-09)
    Traceback (most recent call last):
        ...
    ValueError
    """
    __tracebackhide__ = True
    x = float(x)
    y = float(y)

    if not isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol):
        _raise(handle_with, message)


@switch
def check_if_paths_exist(
    paths: Union[PathType, abc.Iterable[PathType]],
    handle_with: type = FileNotFoundError,
    message: Optional[str] = None,
    execution_mode: str = "raise",
) -> Union[None, Tuple[Optional[Any], List[str]]]:
    """Check if a path or paths exist.

    If it does not, either raise (or return) an exception or issue (or return)
    a warning.

    Args:
        paths (str, pathlib.Path, abc.Sequence[str or pathlib.Path]): path or paths
            to validate
        handle_with (type): type of exception or warning to be raised/returned
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.
        execution_mode (str): defines what happens if not all the paths exist
            May take one of the following values:
                - 'raise': exception/warning will be raised
                - 'return': function will return information about the errors

    Returns:
        None, if execution_mode is 'raise' and check succeeded
        A tuple, if execution_mode is 'return'. The tuple has two elements:
            - an instance of the type provided by the handle_with parameter
            - a list of the non-existing paths

    Raises:
        Exception of the type provided by the handle_with parameter,
        FileNotFoundError by default (unless handle_with is a warning).

    >>> import os
    >>> check_if_paths_exist('Q:/Op/Oop/')
    Traceback (most recent call last):
        ...
    FileNotFoundError
    >>> check_if_paths_exist(os.listdir()[0])
    >>> check_if_paths_exist(Path(os.listdir()[0]))
    >>> check_if_paths_exist(os.listdir())

    >>> check_if_paths_exist('Q:/Op/Oop', execution_mode='return')
    (FileNotFoundError(), ['Q:/Op/Oop'])
    >>> check_if_paths_exist(os.listdir()[0], execution_mode='return')
    (None, [])
    >>> check_if_paths_exist(os.listdir(), execution_mode='return')
    (None, [])

    To issue a warning, do the following (we'll catch the warning):
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_if_paths_exist('Q:/Op/Oop', handle_with=Warning)
    >>> check_if_paths_exist('Q:/Op/Oop',
    ...    execution_mode='return',
    ...    handle_with=Warning)
    (Warning(), ['Q:/Op/Oop'])
    >>> check_if_paths_exist('Q:/Op/Oop',
    ...    execution_mode='return',
    ...    handle_with=Warning,
    ...    message='Attempt to use a non-existing path')
    (Warning('Attempt to use a non-existing path'), ['Q:/Op/Oop'])
    """
    __tracebackhide__ = True
    if execution_mode not in ("raise", "return"):
        _raise(ValueError, "execution_mode must be either 'raise' or 'return'")

    is_allowed_type = isinstance(paths, (str, Path)) or (
        isinstance(paths, abc.Iterable)
        and all(isinstance(path, (str, Path)) for path in paths)
    )

    if not is_allowed_type:
        _raise(
            TypeError,
            "Argument paths must be string" " or pathlib.Path or iterable thereof",
        )

    error = None

    if isinstance(paths, (str, Path)):
        paths = (paths,)

    non_existing_paths = [str(path) for path in paths if not Path(path).exists()]

    if non_existing_paths:
        if execution_mode == "raise":
            _raise(handle_with, message)
        elif execution_mode == "return":
            if message:
                error = handle_with(str(message))
            else:
                error = handle_with()

    if execution_mode == "return":
        return error, non_existing_paths
    return None


@switch
def check_comparison(
    item_1: Any,
    operator: abc.Callable,
    item_2: Any,
    handle_with: type = ValueError,
    message: Optional[str] = None,
) -> None:
    """Check if a comparison of two items is true.

    Args:
        item_1: the first item to compare
        operator: one of the functions returned by get_possible_operators()
        item_2: the second item to compare
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.

    Returns:
        None, if check succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter, ValueError
        by default (unless handle_with is a warning).

    >>> check_comparison(2, lt, 2)
    Traceback (most recent call last):
        ...
    ValueError
    >>> check_comparison(2, eq, 2)
    >>> check_comparison(2, ge, 1.1)
    >>> check_comparison('One text', lt, 'one text')
    >>> check_comparison('One text', lt, 'another text')
    >>> check_comparison('one text', lt, 'another text')
    Traceback (most recent call last):
        ...
    ValueError

    You can use a dedicated ComparisonError class (defined in this module):
    >>> check_comparison('one text', lt, 'another text',
    ...                  handle_with=ComparisonError,
    ...                  message='Not less!')
    Traceback (most recent call last):
        ...
    ComparisonError: Not less!

    To issue a warning, do the following:
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_comparison('one text', lt, 'another text',
    ...         handle_with=Warning,
    ...         message='Not less!')
    ...     assert_if("Not less" in str(w[-1].message))
    """
    __tracebackhide__ = True
    if not operator(item_1, item_2):
        _raise(handle_with, message)


@switch
def check_all_ifs(
    *args: Tuple[abc.Callable, Tuple]
) -> dict:  # abc.Mapping[str, Union[type, Warning, bool]]:
    """Check all multiple conditions and return all checks.

    Args:
        args: tuples of the form (check_function, arguments), where:
            - arguments is a tuple of arguments to be passed to a
              check_function
            - check_function is any of the check functions from this module
            (that is, any of the functions starting off with check_).

    Returns:
        A dict of the following structure:
            {'1: check_if': True, '2: check_if': True}
            This means that two checks were run, both using check_if, and
            both returned True confirmation status (so no exception was
            raised). In case of an exception raised, the resulting dict has the
            following structure:
            {'1: check_if': True, '2: check_if_not': AssertionError()}
            if a failure message was not provided, and otherwise
            {'1: check_if': True, '2: check_if_not': AssertionError('Wrong')}
            where 'Wrong' was the message provided as the argument.

    Use this function if you want to check a list of multiple conditions and
    catch all the errors (and messages) - it does not behave like the other
    functions in the module, since it returns the results of the checks, instead
    of raising an exception (or issuing a warning).

    >>> check_all_ifs(
    ...    (check_if, 2 > 1),
    ...    (check_if, 'a' == 'a')
    ...    )
    {'1: check_if': True, '2: check_if': True}
    >>> check_all_ifs(
    ...    (check_if, 2 > 1),
    ...    (check_if_not, 'a' == 'a')
    ...    )
    {'1: check_if': True, '2: check_if_not': AssertionError()}
    >>> check_all_ifs( # doctest: +ELLIPSIS
    ...    (check_if, 2 > 1),
    ...    (check_if_not, 'a' == 'a', ValueError, 'Wrong!')
    ...    )
    {'1: check_if': True, '2: check_if_not': ValueError(\'Wrong!...

    You can also use this function with warnings:
    >>> check_all_ifs(
    ...    (check_if, 2 > 1),
    ...    (check_if_not, 'a' == 'a', Warning, 'It might be wrong!')
    ...    )
    {'1: check_if': True, '2: check_if_not': Warning('It might be wrong!')}
    """
    __tracebackhide__ = True
    check_length(
        args,
        0,
        operator=gt,
        handle_with=ValueError,
        message="Provide at least one condition.",
    )
    tuple_error_message = (
        "Provide all function calls as tuples in the form of " "(check_function, *args)"
    )
    for arg in args:
        check_type(arg, tuple, message=tuple_error_message)

    results_of_checks = dict()

    for i, this_check in enumerate(args):
        func, *arguments = this_check
        run_this_check = True
        try:
            with warnings.catch_warnings(record=True) as this_warn:
                func(*arguments)
            if this_warn:
                run_this_check = this_warn[-1].message  # type: ignore
        except Exception as e:
            run_this_check = e  # type: ignore

        results_of_checks[f"{i + 1}: {func.__name__}"] = run_this_check

    return results_of_checks


@switch
def check_argument(
    argument: T,
    argument_name: Optional[str] = None,
    expected_type: Union[type, abc.Sequence[type], None] = None,
    expected_choices: Optional[abc.Sequence[T]] = None,
    expected_length: Optional[int] = None,
    handle_with: type = ArgumentValueError,
    message: Optional[str] = None,
    **kwargs: Any,
) -> None:
    """Check if the user provided a correct argument value.

    Args:
        argument: argument value to be validated
        argument_name (str): original name of the argument in the calling
            function. If argument_name is not defined, the error messages will
            not include the name of the argument, but will instead only report
            the default text 'argument'
        expected_type (type, Iterable[type]): the expected type of the item
        expected_choices (Iterable): a list of acceptable values of argument
        expected_length (int): the expected length of the item
        handle_with (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.
        **kwargs: additional arguments passed to check_length (i.e.,
            operator=eq and assign_length_to_others)

    Returns:
        None, if checks succeeded.

    Raises:
        Exception of the type provided by the handle_with parameter,
        ArgumentValueError by default (unless handle_with is a warning).

    You can use this function to check whether an argument's value meets
    various conditions. This is an alternative approach to independent checking
    of these conditions using separate easycheck functions.

    The function performs lazy checking, meaning that it first checks the
    instance (if provided), then choices (if provided), and then the expected
    length (if provided). These checks must not raise a built-in error, because
    this error would be raised before the check is performed.

    >>> check_argument(
    ...    [1, 2, 3], 'x',
    ...    expected_type=tuple,
    ...    expected_length=3
    ...    )
    Traceback (most recent call last):
        ...
    ArgumentValueError: Incorrect type of x; valid type(s): <class 'tuple'>

    The expected_choices argument helps you check whether the user provided
    a valid value for the argument:
    >>> def foo(x):
    ...    check_argument(x, 'x', expected_choices=('first choice',
    ...                                             'second_choice'))
    ...    # whatever foo is doing...
    >>> foo('first choice')
    >>> foo('no choice') # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ArgumentValueError: x's value, no choice, is not among valid ...

    >>> x = 2.0
    >>> check_argument(
    ...    x, 'x',
    ...    expected_type=int) # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ArgumentValueError: Incorrect type of x; valid ... <class 'int'>

    This is how you can check exceptions and errors provided as arguments:
    >>> check_argument(TypeError, 'error_arg', expected_type=type)
    >>> check_argument(TypeError(), 'error_arg', expected_type=Exception)

    You can also issue a warning instead of raising an exception:
    >>> with warnings.catch_warnings(record=True) as w:
    ...     check_argument(
    ...         x, 'x',
    ...         expected_type=int,
    ...         handle_with=Warning,
    ...         message="Incorrect argument's value")
    ...     assert_if("Incorrect argument's value" in str(w[-1].message))
    """
    __tracebackhide__ = True
    if all(item is None for item in (expected_type, expected_choices, expected_length)):
        raise ValueError(
            "check_argument() requires at least one condition to be checked"
        )

    if argument_name is None:
        argument_name = "argument"
    check_type(
        argument_name,
        str,
        handle_with=handle_with,
        message="argument_name must be string",
    )

    if expected_type is not None:
        instance_message = (
            message
            or f"Incorrect type of {argument_name}; valid type(s):" f" {expected_type}"
        )
        check_type(
            item=argument,
            expected_type=expected_type,
            handle_with=handle_with,
            message=instance_message,
        )
    if expected_choices is not None:
        choices_message = (
            message
            or f"{argument_name}'s value, {argument}, "
            f"is not among valid values: {expected_choices}."
        )
        if argument not in expected_choices:
            _raise(handle_with, choices_message)
    if expected_length is not None:
        length_message = (
            message
            or f"Unexpected length of {argument_name}" f" (should be {expected_length})"
        )
        check_length(
            item=argument,
            expected_length=expected_length,
            handle_with=handle_with,
            message=length_message,
            **kwargs,
        )


@switch
def catch_check(
    check_function: abc.Callable, *args: Any, **kwargs: Any
):  # type: ignore
    """Catch an exception or warning raised/issued by a easycheck function.

    Warning: Be aware that catch_check() is a relatively slow function
             compared to most other easycheck functions.

    Args:
        check_function: function to call (one of the public easycheck functions)
        args: positional arguments to pass on to check_function
        kwargs: keyword arguments to pass on to check_function

    Returns:
        Exception or warning instance

    Raises:
        TypeError, if check_function is not a callable
        ArgumentValueError, if check_function is not one of the allowed functions
        ValueError, if an attempt to call another error-catching function is
            detected

    Most easycheck functions return None when the check is successful, and
    otherwise either raise an exception or issue a warning. You can use this
    function to change this behavior: the check_function will still return None
    when all conditions pass, but instead of raising an exception or issuing
    a warning in case of problems, it will return the specified exception
    or warning.

    >>> catch_check(check_if, 2==2)
    >>> catch_check(check_if, 2>2)
    AssertionError()
    >>> my_check = catch_check(check_if, 2>2, ValueError)
    >>> my_check
    ValueError()
    >>> type(my_check)
    <class 'ValueError'>
    >>> check_type(my_check, ValueError)
    >>> raise(my_check)
    Traceback (most recent call last):
        ...
    ValueError
    >>> print(my_check)
    <BLANKLINE>
    >>> catch_check(check_if, condition=2>2, handle_with=ValueError)
    ValueError()
    >>> catch_check(check_length, [2, 2], 3, message='')
    LengthError()
    >>> my_check = catch_check(
    ...    check_type, 25, float, ValueError, 'This is no float!')
    >>> my_check # doctest: +ELLIPSIS
    ValueError('This is no float!'...
    >>> print(str(my_check))
    This is no float!
    >>> my_check = catch_check(check_type, 'a', int)
    >>> my_check
    TypeError()
    >>> raise(my_check)
    Traceback (most recent call last):
        ...
    TypeError

    You can also catch warnings:
    >>> catch_check(check_if, condition=2>2, handle_with=Warning)
    UserWarning(<class 'Warning'>)
    >>> catch_check(check_if,
    ...    condition=2>2,
    ...    handle_with=UserWarning,
    ...    message='Beware of this problem')
    UserWarning('Beware of this problem')
    """
    __tracebackhide__ = True
    check_if(
        check_function in ALL_FUNCTIONS,
        handle_with=TypeError,
        message=(f"{check_function} does not seem to be a easycheck function"),
    )
    check_if_not(
        check_function == check_all_ifs,
        handle_with=ValueError,
        message=(
            "Do not use catch_check for check_all_ifs() "
            "because it itself returns its checks."
        ),
    )
    paths_condition = check_function == check_if_paths_exist and (
        "return" in args or kwargs.get("execution_mode", "") == "return"
    )
    check_if_not(
        paths_condition,
        handle_with=ValueError,
        message=(
            "Do not use catch_check for check_if_paths_exist() "
            'with execution_mode="return" because it itself '
            "returns its checks."
        ),
    )
    check_argument(
        argument=check_function,
        argument_name=check_function.__name__,
        expected_choices=(
            check_if,
            assert_if,
            check_if_not,
            assert_if_not,
            check_if_isclose,
            assert_if_isclose,
            check_if_in_limits,
            assert_if_in_limits,
            check_argument,
            check_comparison,
            check_if_paths_exist,
            assert_paths,
            check_type,
            assert_type,
            check_length,
            assert_length,
        ),
        handle_with=ArgumentValueError,
        message=(
            f"{check_function.__name__} is not"
            " among acceptable valid easycheck functions"
        ),
    )

    try:
        with warnings.catch_warnings(record=True) as possible_warn:
            check_function(*args, **kwargs)
        if possible_warn:
            return possible_warn[-1].message
    except Exception as e:
        return e


@switch
def _raise(error, message=None):
    """Raise exception or issue a warning, with or without message.

    Args:
        error (type): the type of exception or warning to be raised
        message (str): a text to use as the exception/warning message.
            Defaults to None, which means using no message for built-in
            exceptions/warnings, and the docstrings of the exception/warning
            class as a message for custom exceptions.

    Raises:
        Exception of the type provided by the error parameter
        TypeError if error parameter is not an exception or warning class

    Raising exceptions:
    >>> _raise(ValueError)
    Traceback (most recent call last):
       ....
    ValueError
    >>> _raise(TypeError)
    Traceback (most recent call last):
       ....
    TypeError
    >>> _raise(TypeError, 'Incorrect type')
    Traceback (most recent call last):
       ....
    TypeError: Incorrect type

    Issuing warnings (we will catch them in order to see what they contain):
    >>> with warnings.catch_warnings(record=True) as w:
    ...    _raise(Warning)
    ...    assert_if(issubclass(w[-1].category, Warning))
    ...    assert_if('Warning' in str(w[-1].message))
    >>> with warnings.catch_warnings(record=True) as w:
    ...    _raise(Warning, 'Watch out! Something might be wrong.')
    ...    assert_if('Watch out!' in str(w[-1].message))
    """
    __tracebackhide__ = True
    if not isinstance(error, type) or not issubclass(error, Exception):
        raise TypeError("The error argument must be an exception or a warning")

    if message is None:
        # Use docstring as a message only for custom exceptions.
        if error.__name__ not in dir(builtins):
            message = error.__doc__
    elif isinstance(message, str):
        if not message:
            # Instead of passing an empty string,
            # better to pass None as message.
            message = None
        else:
            message = message
    else:
        raise TypeError("Argument message must be either None or string")

    if issubclass(error, Warning):
        if message:
            warnings.warn(message, error)
        else:
            warnings.warn(error)
    elif issubclass(error, Exception):
        if message:
            raise error(message)
        else:
            raise error


def get_possible_operators() -> Tuple[abc.Callable, ...]:
    """Provide a list of possible operators to be used in easycheck functions.

    All of these operators come from the operator module, but not all operators
    from this module are allowed.

    >>> operators = get_possible_operators()
    >>> type(operators[0])
    <class 'builtin_function_or_method'>
    >>> len(operators)
    8
    """
    return eq, le, lt, gt, ge, ne, is_, is_not


# Aliases to be used for testing. Beware not to use warnings with them.


def make_it_true_assertion(func: abc.Callable) -> abc.Callable:
    @wraps(func)
    def assert_func(*args: Any, **kwargs: Any) -> Any:
        if __debug__:
            return func(*args, **kwargs)

    return assert_func


@switch
@make_it_true_assertion
def assert_if(*args: Any, handle_with: type = AssertionError, **kwargs: Any) -> None:
    return check_if(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_if_not(
    *args: Any, handle_with: type = AssertionError, **kwargs: Any
) -> None:
    return check_if_not(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_if_in_limits(
    *args: Any, handle_with: type = AssertionError, **kwargs: Any
) -> None:
    return check_if_in_limits(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_length(
    *args: Any, handle_with: type = AssertionError, **kwargs: Any
) -> None:
    return check_length(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_type(*args: Any, handle_with: type = AssertionError, **kwargs: Any) -> None:
    return check_type(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_paths(*args: Any, handle_with: type = AssertionError, **kwargs: Any) -> None:
    return check_if_paths_exist(*args, handle_with=handle_with, **kwargs)


@switch
@make_it_true_assertion
def assert_if_isclose(
    *args: Any, handle_with: type = AssertionError, **kwargs: Any
) -> None:
    return check_if_isclose(*args, handle_with=handle_with, **kwargs)


# Alias to ensure backward compatibility

check_instance = check_type
assert_instance = assert_type


ALL_FUNCTIONS = (
    assert_if,
    assert_if_isclose,
    assert_if_not,
    assert_if_in_limits,
    assert_instance,
    assert_length,
    assert_paths,
    assert_type,
    catch_check,
    check_all_ifs,
    check_argument,
    check_comparison,
    check_if,
    check_if_isclose,
    check_if_in_limits,
    check_if_not,
    check_if_paths_exist,
    check_instance,
    check_length,
    check_type,
)


if __name__ == "__main__":
    import doctest

    flags = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    doctest.testmod(optionflags=flags)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

extras_requirements = {
    "dev": ["pytest", "wheel", "black"],
}

setuptools.setup(
    name="easycheck",
    version="0.10.1",
    author="Nyggus, Ke Boan & Darsoo",
    author_email="nyggus@gmail.com",
    license="MIT",
    description="A tool for checking conditions in Python",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/nyggus/easycheck",
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
    extras_require=extras_requirements,
)



from .easycheck import (
    # functions
    check_if,
    check_if_not,
    check_type,
    check_if_paths_exist,
    check_length,
    check_if_isclose,
    check_all_ifs,
    check_argument,
    check_comparison,
    catch_check,
    # Exception classes
    ComparisonError,
    ArgumentValueError,
    LengthError,
    NotCloseEnoughError,
    # Testing aliases
    assert_if,
    assert_if_not,
    assert_length,
    assert_if_isclose,
    assert_type,
    assert_paths,
    # Backward-compatibility aliases
    check_instance,
    assert_instance,
)

from operator import eq, le, lt, gt, ge, ne, is_, is_not


import decimal
import fractions
import os
import pytest
import warnings
from collections.abc import Generator
from operator import eq, le, lt, gt, ge, ne, is_, is_not
from pathlib import Path
from unittest import mock

from easycheck.easycheck import (
    check_if,
    assert_if,
    check_if_not,
    assert_if_not,
    check_if_in_limits,
    assert_if_in_limits,
    check_type,
    assert_type,
    check_if_isclose,
    assert_if_isclose,
    check_if_paths_exist,
    assert_paths,
    check_length,
    assert_length,
    check_all_ifs,
    check_argument,
    check_comparison,
    catch_check,
    LimitError,
    ComparisonError,
    ArgumentValueError,
    LengthError,
    NotCloseEnoughError,
    get_possible_operators,
    _raise,
)


def test_check_if_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_if()
    assert check_if(True) is None
    with pytest.raises(AssertionError):
        check_if(False)
    with pytest.raises(
        TypeError, match="The error argument must be an exception or a warning"
    ):
        check_if(False, 1)
    with pytest.raises(
        TypeError, match="message must be either None or string"
    ):
        check_if(0, ValueError, 1)
    with pytest.raises(TypeError, match="takes from 1 to 3 positional"):
        check_if(1, 1, 1, 1)


def test_check_if_positive():
    assert check_if(2 > 1) is None
    assert check_if(2 > 1, Warning) is None
    assert check_if(0 == 0) is None
    assert check_if(0 == 0, Warning) is None
    assert check_if(None is None) is None
    assert check_if(None is None, Warning) is None


def test_check_if_negative():
    with pytest.raises(AssertionError):
        check_if(2 < 1)
    with pytest.raises(ValueError):
        check_if(2 < 1, handle_with=ValueError)
    with pytest.raises(ValueError, match="incorrect value"):
        check_if(2 < 1, handle_with=ValueError, message="incorrect value")


def test_switched_off_checks_exceptions():
    with mock.patch.dict(os.environ, {"EASYCHECK_RUN": "0"}):
        assert check_if(2 < 1) is None
        assert check_if_not(1 == 1) is None
        assert check_if_in_limits(1, 3, 5) is None
        assert check_length(10, 3) is None
        assert check_if_isclose(1.12, 1.123, abs_tol=0.0005) is None
        assert check_type(True, (str, complex)) is None
        assert catch_check(check_if, 2 == 2) is None
        assert check_comparison(3, eq, 2) is None
        assert check_if_paths_exist("__file__", execution_mode="buuu") is None
        assert check_if_paths_exist("Z:/Op/Oop") is None
        assert check_argument(50, "my_arg", expected_type=str) is None
        multiple_check = check_all_ifs(
            (check_if, 2 > 1), (check_if, "a" != "a")
        )
        assert multiple_check is None


def test_switched_off_checks_warnings():
    with mock.patch.dict(os.environ, {"EASYCHECK_RUN": "0"}):
        with warnings.catch_warnings(record=True) as issues:
            check_if(2 < 1, Warning)
        check_if_not(1 == 1, Warning)
        check_if_in_limits(1, 3, 5, handle_with=Warning())
        check_length(10, 3, handle_with=Warning)
        check_type(True, (str, complex), handle_with=Warning)
        catch_check(check_if, 2 == 2, handle_with=Warning)
        check_comparison(3, eq, 2, handle_with=Warning)
        check_if_paths_exist(
            "__file__", execution_mode="buuu", handle_with=Warning
        )
        check_if_paths_exist("Z:/Op/Oop", handle_with=Warning)
        check_argument(50, "my_arg", expected_type=str, handle_with=Warning)
        multiple_check = check_all_ifs(
            (check_if, 2 > 1, Warning), (check_if, "a" != "a", Warning)
        )
        multiple_check is None
    assert not issues


def test_check_if_negative_warnings():
    with warnings.catch_warnings(record=True) as w:
        check_if(2 < 1, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_if(2 < 1, UserWarning, "This is a testing warning")
        assert issubclass(w[-1].category, Warning)
        assert "This is a testing warning" in str(w[-1].message)


def test_check_if_not_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_if_not()
    assert check_if_not(False) is None
    with pytest.raises(AssertionError):
        check_if_not(True)
    with pytest.raises(
        TypeError, match="he error argument must be an exception or a warning"
    ):
        check_if_not(1, 1)
    with pytest.raises(
        TypeError, match="message must be either None or string"
    ):
        check_if_not(1, ValueError, 1)
    with pytest.raises(TypeError, match="takes from 1 to 3 positional"):
        check_if_not(1, 1, 1, 1)


def test_check_if_not_positive():
    assert check_if_not(2 < 1) is None
    assert check_if_not(2 < 1, Warning) is None
    assert check_if_not("a" == "A") is None
    assert check_if_not("a" == "A", Warning) is None


def test_check_if_not_negative():
    with pytest.raises(AssertionError):
        check_if_not(2 > 1)
    with pytest.raises(ValueError):
        check_if_not(2 > 1, handle_with=ValueError)
    with pytest.raises(ValueError, match="incorrect value"):
        check_if_not(2 > 1, handle_with=ValueError, message="incorrect value")


def test_check_if_not_negative_warnings():
    with warnings.catch_warnings(record=True) as w:
        check_if_not(2 > 1, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_if_not(2 > 1, UserWarning, "This is a testing warning")
        assert issubclass(w[-1].category, Warning)
        assert "This is a testing warning" in str(w[-1].message)


def test_check_if_in_limits():
    assert check_if_in_limits(3, 1, 5) is None
    assert check_if_in_limits(3, 1, 5, handle_with=Warning) is None
    assert check_if_in_limits(3, 1) is None
    assert check_if_in_limits(3, 1, handle_with=Warning) is None
    assert check_if_in_limits(3, 3, include_equal=True) is None
    assert (
        check_if_in_limits(3, 3, include_equal=True, handle_with=Warning)
        is None
    )
    assert check_if_in_limits(3, 1, include_equal=False) is None
    assert (
        check_if_in_limits(3, 1, include_equal=False, handle_with=Warning)
        is None
    )
    assert check_if_in_limits(3, upper_limit=5) is None
    assert check_if_in_limits(3, upper_limit=5, handle_with=Warning) is None
    assert check_if_in_limits(3, lower_limit=3) is None
    assert check_if_in_limits(3, lower_limit=3, handle_with=Warning) is None
    assert check_if_in_limits(3, upper_limit=3) is None
    assert check_if_in_limits(3, upper_limit=3, handle_with=Warning) is None

    assert check_if_in_limits(0, 0, 0) is None
    assert check_if_in_limits(0, 0, 0, handle_with=Warning) is None
    assert check_if_in_limits(3.0, 1.0, 5.0) is None
    assert check_if_in_limits(3.0, 1.0, 5.0, handle_with=Warning) is None
    assert check_if_in_limits(3.0, 1, 5) is None
    assert check_if_in_limits(3.0, 1, 5, handle_with=Warning) is None
    assert check_if_in_limits(3, 1.0, 5.0) is None
    assert check_if_in_limits(3, 1.0, 5.0, handle_with=Warning) is None
    assert check_if_in_limits(-1, -3, 3) is None
    assert check_if_in_limits(-1, -3, 3, handle_with=Warning) is None
    assert check_if_in_limits(0, -1000.0, 1000.0) is None
    assert check_if_in_limits(0, -1000.0, 1000.0, handle_with=Warning) is None
    assert check_if_in_limits(0.0005, 0.0004, 0.0006) is None
    assert (
        check_if_in_limits(0.0005, 0.0004, 0.0006, handle_with=Warning) is None
    )
    assert check_if_in_limits(1000.0, 1000.0, 1000.0) is None
    assert (
        check_if_in_limits(1000.0, 1000.0, 1000.0, handle_with=Warning) is None
    )


def test_check_if_in_limits_negative():
    with pytest.raises(LimitError):
        check_if_in_limits(1, 3, 5)
    with pytest.raises(LimitError):
        check_if_in_limits(5, 1, 3)
    with pytest.raises(LimitError):
        check_if_in_limits(0, 1, 1000)
    with pytest.raises(LimitError):
        check_if_in_limits(3, 1, 3, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(3, 3, 5, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(0, 0, 0, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(-5, 0, 1000, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(1.0, 3.0, 5.0)
    with pytest.raises(LimitError):
        check_if_in_limits(0.0000001, 0.0000002, 0.0000003)
    with pytest.raises(LimitError):
        check_if_in_limits(5.0, 1.0, 3.0)
    with pytest.raises(LimitError):
        check_if_in_limits(3.05, 3.05, 5.05, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(5.00005, 1.2345, 5.00005, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(0.00, 0.00, 0.00, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(-1.05, -0.05, 0.05, include_equal=False)
    with pytest.raises(LimitError):
        check_if_in_limits(-1000.00, -999.999, -999.998, include_equal=False)

    with pytest.raises(TypeError):
        check_if_in_limits(None)


def test_check_length_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_length("tomato soup is good")
    with pytest.raises(TypeError, match="'int' object is not callable"):
        check_length([1], 1, operator=1)


def test_check_length_positive():
    assert check_length(["string"], 1) is None
    assert check_length(["string"], 1, handle_with=Warning) is None
    assert check_length("string", 6) is None
    assert check_length("string", 6, handle_with=Warning) is None
    assert check_length([1, 2], 2) is None
    assert check_length([1, 2], 2, handle_with=Warning) is None
    assert check_length(range(0, 3), 3) is None
    assert check_length(range(0, 3), 3, handle_with=Warning) is None
    assert check_length(10, 1, assign_length_to_others=True) is None
    assert (
        check_length(10, 1, assign_length_to_others=True, handle_with=Warning)
        is None
    )
    assert (
        check_length(
            decimal.Decimal("3.55634"), 1, assign_length_to_others=True
        )
        is None
    )
    assert (
        check_length(
            fractions.Fraction(3, 55), 1, assign_length_to_others=True
        )
        is None
    )


def test_check_length_negative():
    with pytest.raises(TypeError):
        check_length(len(i for i in range(3)), 3)
    with pytest.raises(TypeError):
        check_length(None)
    with pytest.raises(TypeError, match="object of type 'int' has no len()"):
        check_length(10, 1)
    with pytest.raises(TypeError, match="'decimal.Decimal' has no len()"):
        check_length(
            decimal.Decimal("3.55634"), 1, assign_length_to_others=False
        )
    with pytest.raises(TypeError, match="'Fraction' has no len()"):
        check_length(
            fractions.Fraction(3, 55), 1, assign_length_to_others=False
        )


def test_check_length_negative_warnings():
    with warnings.catch_warnings(record=True) as w:
        check_length(
            [1, 2],
            expected_length=1,
            handle_with=Warning,
            message="This is a testing warning",
        )
        assert "This is a testing warning" in str(w[-1].message)


def test_check_if_isclose_edge_cases():
    with pytest.raises(ValueError, match="tolerances must be non-negative"):
        check_if_isclose(1.1, 1.2, abs_tol=-1)
    with pytest.raises(ValueError, match="tolerances must be non-negative"):
        check_if_isclose(1.1, 1.2, rel_tol=-1)
    with pytest.raises(TypeError, match="must be real number, not str"):
        check_if_isclose(1.1, 1.2, rel_tol=".1")
    with pytest.raises(TypeError, match="must be real number, not str"):
        check_if_isclose(1.1, 1.2, abs_tol=".1")
    with pytest.raises(
        ValueError, match="could not convert string to float: '1,1'"
    ):
        check_if_isclose("1,1", "1.2", abs_tol=".1")
    with pytest.raises(
        TypeError,
        match=("positional-only arguments passed" " as keyword arguments"),
    ):
        check_if_isclose(x="1.1", y="1.2")


def test_check_if_isclose_positive():
    assert check_if_isclose(1.12, 1.12, abs_tol=0.01) is None
    assert check_if_isclose(1.12, 1.123, abs_tol=0.05) is None
    assert check_if_isclose(1.12, 1.123, rel_tol=0.01) is None
    assert check_if_isclose(1.12, 1.123, rel_tol=0.05) is None

    assert check_if_isclose("1.12", 1.12, abs_tol=0.01) is None
    assert check_if_isclose(1.12, "1.12", abs_tol=0.01) is None
    assert check_if_isclose("1.12", "1.12", abs_tol=0.01) is None
    assert check_if_isclose(" 1.12 ", 1.12, abs_tol=0.01) is None
    assert check_if_isclose(" 1.12 ", "\t1.12\n", abs_tol=0.01) is None
    assert (
        check_if_isclose(
            " 1.12 ", "\t     1.12   \n  \n   \n\n\n", abs_tol=0.01
        )
        is None
    )

    assert (
        check_if_isclose(1.12, 1.123, rel_tol=0.05, handle_with=ValueError)
        is None
    )

    # any of the two check (abs_tol or rel_tol) is enough
    # for the test to pass:
    assert check_if_isclose(1.12, 1.123, rel_tol=0.05, abs_tol=0.05) is None

    check_if_isclose(1.12, 1.123, rel_tol=0.000005, abs_tol=0.05)


def test_check_if_isclose_negative():
    with pytest.raises(NotCloseEnoughError):
        check_if_isclose(1.12, 1.123, abs_tol=0.0005)
    with pytest.raises(NotCloseEnoughError):
        check_if_isclose(
            1.12, 1.123, message="Not close", rel_tol=0, abs_tol=0.0005
        )
    with pytest.raises(NotCloseEnoughError):
        check_if_isclose(1.12, 1.123, rel_tol=0.0005)


def test_check_type_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_type("tomato soup is good")


def test_check_type_positive():
    assert check_type(["string"], list) is None
    assert check_type(["string"], list, handle_with=Warning) is None
    assert check_type("string", str) is None
    assert check_type("string", str, handle_with=Warning) is None
    assert check_type((1, 2), tuple) is None
    assert check_type((1, 2), tuple, handle_with=Warning) is None
    assert check_type((1, 2), [tuple, list]) is None
    assert check_type((1, 2), [tuple, list], handle_with=Warning) is None
    assert check_type((1, 2), [list, tuple]) is None
    assert check_type((1, 2), {tuple, list}, handle_with=Warning) is None
    assert (
        check_type((1, 2), (tuple, list), message="Neither tuple nor list")
        is None
    )
    assert (
        check_type(
            (1, 2),
            (tuple, list),
            message="Neither tuple nor list",
            handle_with=Warning,
        )
        is None
    )
    assert check_type((i for i in range(3)), Generator) is None
    assert (
        check_type((i for i in range(3)), Generator, handle_with=Warning)
        is None
    )
    assert check_type(None, (int, None)) is None
    assert check_type(None, (int, None), handle_with=Warning) is None
    assert check_type(20, (int, None)) is None
    assert check_type(20, (int, None), handle_with=Warning) is None
    assert check_type(None, None) is None
    assert check_type(None, None, handle_with=Warning) is None


def test_check_type_negative():
    with pytest.raises(TypeError, match="Neither tuple nor list"):
        check_type("souvenir", (tuple, list), message="Neither tuple nor list")
    with pytest.raises(TypeError):
        check_type("souvenir", [tuple, list])
    with pytest.raises(TypeError):
        check_type("souvenir", {tuple, list})
    with pytest.raises(TypeError):
        check_type(True, (str, complex))
    with pytest.raises(TypeError):
        check_type(20.1, (int, None))
    with pytest.raises(TypeError):
        check_type((i for i in range(3)), tuple)
    with pytest.raises(TypeError, match="This is not tuple"):
        check_type((i for i in range(3)), tuple, message="This is not tuple")
    with pytest.raises(TypeError):
        check_type(10, None)
    with pytest.raises(TypeError):
        check_type("string", None)
    with pytest.raises(TypeError):
        check_type((10, 20), None)
    with pytest.raises(TypeError):
        check_type([10, 20], None)


def test_check_type_negative_warnings():
    with warnings.catch_warnings(record=True) as w:
        check_type(
            "souvenir",
            (tuple, list),
            handle_with=Warning,
            message="This is a testing warning",
        )
        assert "This is a testing warning" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_type(
            True,
            [str, complex],
            handle_with=Warning,
            message="This is a testing warning",
        )
        assert "This is a testing warning" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_type(
            20.1,
            (int, None),
            handle_with=Warning,
            message="This is a testing warning",
        )
        assert "This is a testing warning" in str(w[-1].message)


def test_catch_check_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        catch_check(check=check_type)
    with pytest.raises(TypeError, match="easycheck function"):
        catch_check(1)
    with pytest.raises(
        TypeError, match="does not seem to be a easycheck function"
    ):
        catch_check(sum)


def test_catch_check_if():
    my_check = catch_check(check_if, 2 == 2)
    assert my_check is None

    my_check = catch_check(check_if, 2 == 2, Warning)
    assert my_check is None

    my_check_not = catch_check(check_if, 2 > 2)
    with pytest.raises(AssertionError):
        raise my_check_not

    my_check_not = catch_check(check_if, 2 > 2, UserWarning, "Problem!")
    assert isinstance(my_check_not, Warning)

    my_check_not = catch_check(check_if, 2 > 2, ValueError)
    assert isinstance(my_check_not, ValueError)
    with pytest.raises(ValueError):
        raise my_check_not

    my_check_not = catch_check(check_if, 2 > 2, Warning)
    assert isinstance(my_check_not, Warning)


def test_catch_check_if_not():
    my_check = catch_check(check_if_not, 2 > 2)
    assert my_check is None

    my_check = catch_check(check_if_not, 2 > 2, Warning, "Problem!")
    assert my_check is None

    my_check_not = catch_check(check_if_not, 2 == 2)
    assert isinstance(my_check_not, AssertionError)
    with pytest.raises(AssertionError):
        raise my_check_not

    my_check_not = catch_check(
        check_if_not, 2 == 2, handle_with=Warning, message="Problem!"
    )
    assert isinstance(my_check_not, Warning)


def test_catch_check_if_in_limits():
    my_check = catch_check(check_if_in_limits, 3, 1, 5)
    assert my_check is None

    my_check = catch_check(check_if_in_limits, 3, 1, 5, handle_with=Warning)
    assert my_check is None

    my_check = catch_check(check_if_in_limits, 0, 1, 5, handle_with=Warning)
    assert isinstance(my_check, Warning)

    my_check = catch_check(
        check_if_in_limits,
        0,
        1,
        5,
        handle_with=Warning,
        message="Number out of limits",
    )
    assert isinstance(my_check, Warning)
    assert "Number out of limits" in str(my_check)

    my_check = catch_check(check_if_in_limits, 0, 1, 5)
    assert isinstance(my_check, LimitError)
    with pytest.raises(LimitError):
        raise my_check


def test_catch_check_length():
    my_check = catch_check(check_length, [2, 2], 2)
    assert my_check is None

    my_check = catch_check(
        check_length,
        [2, 2],
        expected_length=2,
        handle_with=Warning,
        message="Length problem",
    )
    assert my_check is None

    my_check_not = catch_check(check_length, [2, 2], 3)
    assert isinstance(my_check_not, LengthError)
    with pytest.raises(LengthError):
        raise my_check_not

    my_check = catch_check(
        check_length,
        [2, 2],
        expected_length=3,
        handle_with=Warning,
        message="Length problem",
    )
    assert isinstance(my_check, Warning)
    assert "Length problem" in str(my_check)


def test_catch_check_type():
    my_check = catch_check(check_type, 25, int)
    assert my_check is None

    my_check = catch_check(check_type, 25, int, Warning, "Instance issue")
    assert my_check is None

    my_check_not = catch_check(
        check_type, 25, float, ValueError, "This is no float!"
    )
    assert isinstance(my_check_not, ValueError)
    with pytest.raises(ValueError, match="This is no float!"):
        raise my_check_not

    my_check = catch_check(check_type, 25, float, Warning, "Instance issue")
    assert isinstance(my_check, Warning)

    my_check = catch_check(check_type, "a", int)
    assert isinstance(my_check, TypeError)
    with pytest.raises(TypeError):
        raise my_check


def test_catch_check_paths_with_return():
    existing_path = os.listdir(".")[0]
    with pytest.raises(ValueError, match='execution_mode="return"'):
        catch_check(
            check_if_paths_exist, paths=existing_path, execution_mode="return"
        )
    with pytest.raises(ValueError, match='execution_mode="return"'):
        catch_check(
            check_if_paths_exist,
            existing_path,
            FileNotFoundError,
            "Path not found",
            "return",
        )


def test_catch_check_paths_one_path():
    existing_path = os.listdir(".")[0]
    my_check = catch_check(check_if_paths_exist, paths=existing_path)
    my_check_path = catch_check(
        check_if_paths_exist, paths=Path(existing_path)
    )
    assert my_check is None
    assert my_check_path is None

    my_check = catch_check(
        check_if_paths_exist,
        paths=existing_path,
        handle_with=Warning,
        message="Path problem",
    )
    assert my_check is None
    my_check_path = catch_check(
        check_if_paths_exist,
        paths=Path(existing_path),
        handle_with=Warning,
        message="Path problem",
    )
    assert my_check_path is None

    non_existing_path = "W:/Op/No_no"
    my_check_not = catch_check(check_if_paths_exist, paths=non_existing_path)
    my_check_not_path = catch_check(
        check_if_paths_exist, paths=Path(non_existing_path)
    )
    assert isinstance(my_check_not, FileNotFoundError)
    with pytest.raises(FileNotFoundError):
        raise my_check_not
    assert isinstance(my_check_not_path, FileNotFoundError)
    with pytest.raises(FileNotFoundError):
        raise my_check_not_path

    my_check_not = catch_check(
        check_if_paths_exist,
        paths=non_existing_path,
        handle_with=Warning,
        message="Path problem",
    )
    assert isinstance(my_check_not, Warning)


def test_catch_check_paths_many_paths():
    existing_paths = os.listdir(".")
    my_check = catch_check(check_if_paths_exist, paths=existing_paths)
    my_check_path = catch_check(
        check_if_paths_exist, paths=[Path(path) for path in existing_paths]
    )
    assert my_check is None
    assert my_check_path is None

    my_check = catch_check(
        check_if_paths_exist,
        paths=existing_paths,
        handle_with=Warning,
        message="Path issue",
    )
    assert my_check is None

    non_existing_paths = ["W:/Op/No_no"] + os.listdir(".")
    my_check_not = catch_check(check_if_paths_exist, paths=non_existing_paths)
    my_check_not_path = catch_check(
        check_if_paths_exist, paths=[Path(path) for path in non_existing_paths]
    )
    assert isinstance(my_check_not, FileNotFoundError)
    with pytest.raises(FileNotFoundError):
        raise my_check_not
    assert isinstance(my_check_not_path, FileNotFoundError)
    with pytest.raises(FileNotFoundError):
        raise my_check_not_path

    my_check_not = catch_check(
        check_if_paths_exist,
        paths=non_existing_paths,
        handle_with=Warning,
        message="Path issue",
    )
    assert isinstance(my_check_not, Warning)


def test_catch_check_if_isclose():
    my_check = catch_check(check_if_isclose, 1.12, 1.12, abs_tol=0.01)
    assert my_check is None

    my_check = catch_check(check_if_isclose, 1.12, 1.123, abs_tol=0.05)
    assert my_check is None

    my_check = catch_check(check_if_isclose, 1.12, 1.123, rel_tol=0.01)
    assert my_check is None

    my_check = catch_check(check_if_isclose, 1.12, 1.123, rel_tol=0.05)
    assert my_check is None

    my_check_not = catch_check(check_if_isclose, 1.12, 1.123, abs_tol=0.0005)
    assert isinstance(my_check_not, NotCloseEnoughError)
    with pytest.raises(NotCloseEnoughError):
        raise my_check_not

    my_check_not = catch_check(
        check_if_isclose,
        1.12,
        1.123,
        message="Not close",
        rel_tol=0,
        abs_tol=0.0005,
    )
    assert isinstance(my_check_not, NotCloseEnoughError)
    with pytest.raises(NotCloseEnoughError):
        raise my_check_not


def test_check_comparison_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_comparison(1, 1)
    with pytest.raises(TypeError, match="unexpected keyword"):
        check_comparison(item1=1, operator=gt, item_2=2)
    with pytest.raises(TypeError, match="unexpected keyword"):
        check_comparison(item_1=1, Operator=gt, item_2=2)
    with pytest.raises(TypeError, match="unexpected keyword"):
        check_comparison(item_1=1, operator=gt, item2=2)


def test_check_comparison_positive():
    assert check_comparison(2, eq, 2) is None
    assert check_comparison(2, eq, 2, Warning) is None
    assert check_comparison(2, ne, 3) is None
    assert check_comparison(2, ne, 3, Warning) is None
    assert check_comparison(2, le, 2) is None
    assert check_comparison(2, le, 2, Warning) is None
    assert check_comparison(2, ge, 2) is None
    assert check_comparison(2, ge, 2, Warning) is None
    assert check_comparison(3, ge, 2) is None
    assert check_comparison(3, ge, 2, Warning) is None
    assert check_comparison(3, gt, 2) is None
    assert check_comparison(3, gt, 2, Warning) is None
    assert check_comparison("One text", lt, "one text") is None
    assert check_comparison("One text", lt, "one text", Warning) is None
    assert check_comparison("One text", lt, "another text") is None
    assert check_comparison("One text", lt, "another text", Warning) is None
    assert check_comparison("abc", is_, "abc") is None
    assert check_comparison("abc", is_, "abc", Warning) is None
    assert check_comparison("abc", is_not, "xyz") is None
    assert check_comparison("abc", is_not, "xyz", Warning) is None


def test_check_comparison_negative():
    with pytest.raises(ValueError):
        check_comparison(3, eq, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(3, eq, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison(2, ne, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(2, ne, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison(2, lt, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(2, lt, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison(2, gt, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(2, gt, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison(3, lt, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(3, lt, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison(3, le, 2)
    with warnings.catch_warnings(record=True) as w:
        check_comparison(3, le, 2, Warning, "This is a testing warning")
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison("one text", lt, "another text")
    with warnings.catch_warnings(record=True) as w:
        check_comparison(
            "one text",
            lt,
            "another text",
            Warning,
            "This is a testing warning",
        )
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison("abc", is_, "xyz")
    with warnings.catch_warnings(record=True) as w:
        check_comparison(
            "abc", is_, "xyz", Warning, "This is a testing warning"
        )
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ValueError):
        check_comparison("abc", is_not, "abc")
    with warnings.catch_warnings(record=True) as w:
        check_comparison(
            "abc", is_not, "abc", Warning, "This is a testing warning"
        )
        assert "This is a testing warning" in str(w[-1].message)

    with pytest.raises(ComparisonError):
        check_comparison(
            "one text", lt, "another text", handle_with=ComparisonError
        )


def test_check_all_ifs_edge_cases():
    with pytest.raises(ValueError, match="at least one condition"):
        check_all_ifs()
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs(1)
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs(True)
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs(1, 1)
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs(1 > 1, 2 > 1)
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs((20 > 10))
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs(check_if(20 > 10)),
    with pytest.raises(TypeError, match="Provide all function calls as"):
        check_all_ifs((check_if, 20 > 10), (check_if(20 > 10)))


def test_check_all_ifs():
    multiple_check_1 = check_all_ifs((check_if, 2 > 1), (check_if, "a" == "a"))
    assert all(multiple_check_1.values())

    multiple_check_2 = check_all_ifs(
        (check_if, 2 > 1), (check_if_not, "a" == "a")
    )
    assert any(
        type(value) == AssertionError for value in multiple_check_2.values()
    )

    multiple_check_3 = check_all_ifs(
        (check_if, 2 > 1), (check_if_not, "a" == "a", ValueError)
    )
    assert any(
        type(value) == ValueError for value in multiple_check_3.values()
    )


def test_check_all_ifs_warnings():
    multiple_check_1 = check_all_ifs(
        (check_if, 2 > 1, Warning), (check_if, "a" == "a", Warning)
    )
    assert all(multiple_check_1.values())

    multiple_check_2 = check_all_ifs(
        (check_if, 2 > 1, Warning), (check_if_not, "a" == "a", Warning)
    )
    assert any(
        isinstance(value, Warning) for value in multiple_check_2.values()
    )


def test_check_if_paths_exist_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_if_paths_exist()
    with pytest.raises(TypeError, match="unexpected keyword"):
        check_if_paths_exist(path="tomato soup is good")
    with pytest.raises(TypeError, match="Argument paths must be string"):
        check_if_paths_exist(20)
    with pytest.raises(TypeError, match="Argument paths must be string"):
        check_if_paths_exist(True)
    with pytest.raises(TypeError, match="Argument paths must be string"):
        check_if_paths_exist(["/some/path", 1, 2])
    with pytest.raises(TypeError, match="Argument paths must be string"):
        check_if_paths_exist(["/some/path", Path("/some/path"), False])
    with pytest.raises(ValueError):
        check_if_paths_exist(os.listdir(".")[0], execution_mode="buuu")


def test_check_if_paths_exist_positive():
    single_path_to_check = os.listdir(".")[0]
    list_of_paths_to_check = os.listdir(".")
    assert check_if_paths_exist(single_path_to_check) is None
    assert check_if_paths_exist(single_path_to_check, Warning) is None
    assert check_if_paths_exist(single_path_to_check, execution_mode="return")
    assert check_if_paths_exist(
        single_path_to_check, handle_with=Warning, execution_mode="return"
    )
    assert check_if_paths_exist(Path(single_path_to_check)) is None

    assert check_if_paths_exist(list_of_paths_to_check) is None
    assert check_if_paths_exist(list_of_paths_to_check, Warning) is None
    assert (
        check_if_paths_exist(Path(path) for path in list_of_paths_to_check)
        is None
    )

    check_result = check_if_paths_exist(
        list_of_paths_to_check, execution_mode="return"
    )
    assert len(check_result) == 2
    assert check_result[0] is None
    assert check_result[1] == []

    check_result = check_if_paths_exist(
        list_of_paths_to_check, handle_with=Warning, execution_mode="return"
    )
    assert len(check_result) == 2
    assert check_result[0] is None
    assert check_result[1] == []


def test_check_if_paths_exist_negative():
    non_existing_path = "Z:/Op/Oop"
    with pytest.raises(ValueError):
        check_if_paths_exist(non_existing_path, execution_mode="buuu")
    with pytest.raises(FileNotFoundError):
        check_if_paths_exist(non_existing_path)
    with pytest.raises(FileNotFoundError):
        check_if_paths_exist(Path(non_existing_path))
    with pytest.raises(FileNotFoundError):
        check_if_paths_exist([non_existing_path] + os.listdir("."))
    with pytest.raises(IOError):
        check_if_paths_exist(non_existing_path, handle_with=IOError)

    check_result = check_if_paths_exist(
        non_existing_path, execution_mode="return"
    )
    assert len(check_result) == 2
    assert type(check_result[0]) == FileNotFoundError
    with pytest.raises(FileNotFoundError):
        raise check_result[0]
    assert check_result[1] == [non_existing_path]

    check_result = check_if_paths_exist(
        os.listdir(".") + [non_existing_path], execution_mode="return"
    )
    assert len(check_result) == 2
    assert type(check_result[0]) == FileNotFoundError
    with pytest.raises(FileNotFoundError):
        raise check_result[0]
    assert check_result[1] == [non_existing_path]


def test_check_if_paths_exist_negative_warnings():
    non_existing_path = "Z:/Op/Oop"
    with warnings.catch_warnings(record=True):
        check_if_paths_exist(non_existing_path, Warning, "Path issue")
    with warnings.catch_warnings(record=True):
        check_if_paths_exist(Path(non_existing_path), Warning, "Path issue")
    with warnings.catch_warnings(record=True):
        check_if_paths_exist(
            [non_existing_path] + os.listdir("."), Warning, "Path issue"
        )

    check_result = check_if_paths_exist(
        non_existing_path,
        handle_with=Warning,
        message="Path issue",
        execution_mode="return",
    )
    assert len(check_result) == 2
    assert type(check_result[0]) == Warning
    assert "Path issue" in str(check_result[0])
    assert check_result[1] == [non_existing_path]

    check_result = check_if_paths_exist(
        [non_existing_path] + os.listdir("."),
        handle_with=Warning,
        message="Path issue",
        execution_mode="return",
    )
    assert len(check_result) == 2
    assert type(check_result[0]) == Warning
    assert "Path issue" in str(check_result[0])
    assert check_result[1] == [non_existing_path]


def test_raise_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        _raise()
    with pytest.raises(TypeError, match="unexpected keyword"):
        _raise(handle_with=TypeError)
    with pytest.raises(TypeError, match="unexpected keyword"):
        _raise(handle_with=TypeError, MEssage="This was an error")
    with pytest.raises(
        TypeError, match="The error argument must be an exception or a warning"
    ):
        _raise(20)
    with pytest.raises(
        TypeError, match="The error argument must be an exception or a warning"
    ):
        _raise("TypeError")
    with pytest.raises(
        TypeError, match="The error argument must be an exception or a warning"
    ):
        _raise(["TypeError"])
    with pytest.raises(
        TypeError, match="The error argument must be an exception or a warning"
    ):
        _raise(NotImplemented)
    with pytest.raises(
        TypeError, match="Argument message must be either None or string"
    ):
        _raise(error=TypeError, message=20)
    with pytest.raises(
        TypeError, match="Argument message must be either None or string"
    ):
        _raise(TypeError, ("This was an error", ""))


def test_raise_exception():
    with pytest.raises(ValueError):
        _raise(ValueError)
    with pytest.raises(TypeError):
        _raise(TypeError)
    with pytest.raises(TypeError, match="Incorrect type"):
        _raise(TypeError, "Incorrect type")


def test_raise_warning():
    with warnings.catch_warnings(record=True):
        _raise(Warning)
    with warnings.catch_warnings(record=True):
        _raise(UserWarning)
    with warnings.catch_warnings(record=True):
        _raise(Warning, message="Problem with something")
    with warnings.catch_warnings(record=True):
        _raise(UserWarning, message="Problem with something")


def test_check_argument_edge_cases():
    with pytest.raises(TypeError, match="required positional argument"):
        check_argument(Argument="x")

    msg = "check_argument() requires at least one condition to be checked"
    with pytest.raises(ValueError) as msg_error:
        check_argument(10)
    assert str(msg_error.value) == msg
    with pytest.raises(ValueError) as msg_error:
        check_argument(10, message="Error!")
    assert str(msg_error.value) == msg
    with pytest.raises(ValueError) as msg_error:
        check_argument("x", 10, handle_with=TypeError)
    assert str(msg_error.value) == msg


def test_check_argument_type():
    def foo(x):
        check_argument(x, "x", expected_type=str)
        pass

    assert foo("one") is None
    with pytest.raises(ArgumentValueError):
        foo(4)
    with pytest.raises(ArgumentValueError):
        foo(("one", "two"))

    assert check_argument(50, "my_arg", expected_type=int) is None
    assert check_argument(50, expected_type=int) is None
    assert check_argument("my_arg", "one", expected_type=str) is None
    assert check_argument("my_arg", expected_type=str) is None

    with pytest.raises(ArgumentValueError, match="my_arg"):
        check_argument(50, "my_arg", expected_type=str)
    with pytest.raises(ArgumentValueError, match="my_arg"):
        check_argument("one", "my_arg", expected_type=int)
    with pytest.raises(ArgumentValueError, match="argument"):
        check_argument(50, expected_type=str)
    with pytest.raises(ArgumentValueError, match="argument"):
        check_argument("one", expected_type=int)


def test_check_argument_type_warning():
    def foo(x):
        check_argument(
            x,
            "x",
            expected_type=str,
            handle_with=Warning,
            message="Incorrect argument?",
        )
        pass

    assert foo("one") is None
    with warnings.catch_warnings(record=True) as w:
        foo(4)
        assert "Incorrect argument" in str(w[-1].message)

    assert (
        check_argument(50, "my_arg", expected_type=int, handle_with=Warning)
        is None
    )
    assert check_argument(50, expected_type=int, handle_with=Warning) is None

    with warnings.catch_warnings(record=True) as w:
        check_argument(50, "my_arg", expected_type=str, handle_with=Warning)
        assert "my_arg" in str(w[-1].message)
        assert "Incorrect type" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_argument("one", "my_arg", expected_type=int, handle_with=Warning)
        assert "my_arg" in str(w[-1].message)
        assert "Incorrect type" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_argument(50, expected_type=str, handle_with=Warning)
        assert "Incorrect type of argument" in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        check_argument("one", expected_type=int, handle_with=Warning)
        assert "Incorrect type of argument" in str(w[-1].message)


def test_check_argument_choices():
    assert (
        check_argument(
            5,
            "my_arg",
            expected_choices=range(10),
        )
        is None
    )
    assert check_argument(5, expected_choices=range(10)) is None

    def foo(x):
        check_argument(x, expected_choices=("first choice", "second choice"))
        pass

    assert foo("first choice") is None
    assert foo("second choice") is None
    with pytest.raises(ArgumentValueError, match="not among valid values"):
        foo("no choice")

    def foo(x):
        check_argument(x, "x", expected_choices=("one", "two"))
        pass

    assert foo("one") is None
    with pytest.raises(
        ArgumentValueError, match="x's value, three, is not among valid values"
    ):
        foo("three")

    def foo(x):
        check_argument(x, expected_choices=("one", "two"))
        pass

    assert foo("one") is None
    with pytest.raises(
        ArgumentValueError,
        match="argument's value, three, is not among valid values",
    ):
        foo("three")


def test_check_argument_choices_warnings():
    assert (
        check_argument(
            5, "my_arg", expected_choices=range(10), handle_with=Warning
        )
        is None
    )
    assert (
        check_argument(5, expected_choices=range(10), handle_with=Warning)
        is None
    )

    def foo(x):
        check_argument(
            x,
            expected_choices=("first choice", "second choice"),
            handle_with=Warning,
        )
        pass

    assert foo("first choice") is None
    assert foo("second choice") is None
    with warnings.catch_warnings(record=True) as w:
        foo("no choice")
        assert "no choice" in str(w[-1].message)
        assert "not among valid values" in str(w[-1].message)


def test_check_argument_length():
    assert (
        check_argument(
            5, "my_arg", expected_length=1, assign_length_to_others=True
        )
        is None
    )
    assert (
        check_argument(5, expected_length=1, assign_length_to_others=True)
        is None
    )

    def foo(x):
        check_argument(x, "x", expected_length=3, assign_length_to_others=True)
        pass

    assert foo([1, 2, 3]) is None
    with pytest.raises(ArgumentValueError):
        foo(1)

    def foo(x):
        check_argument(x, expected_length=3, assign_length_to_others=True)
        pass

    assert foo([1, 2, 3]) is None
    with pytest.raises(ArgumentValueError):
        foo(1)

    def foo(big_x):
        check_argument(
            big_x, "big_x", expected_length=3, assign_length_to_others=True
        )
        pass

    assert foo([1, 2, 3]) is None
    with pytest.raises(ArgumentValueError, match="big_x"):
        foo(1)

    def foo(big_x):
        check_argument(big_x, expected_length=3, assign_length_to_others=True)
        pass

    assert foo([1, 2, 3]) is None
    with pytest.raises(ArgumentValueError, match="argument"):
        foo(1)


def test_check_argument_length_warnings():
    assert (
        check_argument(
            5,
            "my_arg",
            expected_length=1,
            assign_length_to_others=True,
            handle_with=Warning,
        )
        is None
    )
    assert (
        check_argument(
            5,
            expected_length=1,
            assign_length_to_others=True,
            handle_with=Warning,
        )
        is None
    )

    def foo(x):
        check_argument(
            x,
            "x",
            expected_length=3,
            assign_length_to_others=True,
            handle_with=Warning,
        )
        pass

    assert foo([1, 2, 3]) is None
    with warnings.catch_warnings(record=True) as w:
        foo(1)
        assert "length" in str(w[-1].message)

    def foo(x):
        check_argument(
            x,
            expected_length=3,
            assign_length_to_others=True,
            handle_with=Warning,
        )
        pass

    assert foo([1, 2, 3]) is None
    with warnings.catch_warnings(record=True) as w:
        foo(1)
        assert "length" in str(w[-1].message)


def test_check_argument_mix():
    def foo(x):
        check_argument(x, "x", expected_type=int, condition=x % 2 == 0)
        pass

    with pytest.raises(TypeError):
        x = "one"
        check_argument(
            argument=x,
            argument_name="x",
            expected_type=int,
            condition=x % 2 == 0,
        )
    with pytest.raises(TypeError):
        foo("one")


def test_check_argument_mix_warnings():
    def foo(x):
        check_argument(
            x, "x", expected_type=int, expected_length=3, handle_with=Warning
        )
        pass

    with warnings.catch_warnings(record=True) as w:
        foo("one")
        assert "type" in str(w[-1].message)


def test_assert_if():
    assert assert_if(10 > 5) == check_if(10 > 5)
    with pytest.raises(AssertionError):
        assert_if(10 < 5)

def test_assert_if_not():
    assert assert_if_not(10 < 5) == check_if_not(10 < 5)
    with pytest.raises(AssertionError):
        assert_if_not(10 > 5)

def test_assert_if_in_linits():
    assert assert_if_in_limits(3, 1, 5) == check_if_in_limits(3, 1, 5)
    with pytest.raises(AssertionError):
        assert_if_in_limits(1, 3, 5)

def test_assert_type():
    assert assert_type((10, 10), tuple) == check_type((10, 10), tuple)
    with pytest.raises(AssertionError):
        assert_type(10, tuple)

def test_assert_length():
    assert assert_length("str", 3) == check_length("str", 3)
    assert assert_length(5, 1, assign_length_to_others=True) == check_length(
        5, 1, assign_length_to_others=True
    )
    with pytest.raises(TypeError):
        assert_length(5, 3)
    with pytest.raises(AssertionError):                        
        assert_length(5, 3, assign_length_to_others=True)

def test_assert_if_iscolse():
    assert assert_if_isclose(1.12, 1.123, abs_tol=0.05) == check_if_isclose(1.12, 1.123, abs_tol=0.05)
    with pytest.raises(AssertionError):
        assert_if_isclose(1.12, 1.123, abs_tol=0.0005)

def test_assert_paths():
    existing_file = os.listdir(".")[0]
    assert check_if_paths_exist(
        existing_file, execution_mode="return"
    ) == assert_paths(existing_file, execution_mode="return")
    assert (
        check_if_paths_exist("Q:/E/", execution_mode="return")[1]
        == assert_paths("Q:/E/", execution_mode="return")[1]
    )
    with pytest.raises(AssertionError):
        assert_paths("Q:/E/")

class ForTestingErrorWithDoc(Exception):
    """This is error for testing purposes."""


class ForTestingErrorWithoutDoc(Exception):
    ...


def test_message_is_None_exception_with_docstring():
    with pytest.raises(AssertionError, match=""):
        check_if(1 == 2, message=None)
    with pytest.raises(AssertionError, match="Error"):
        check_if(1 == 2, message="Error")
    with pytest.raises(AssertionError):
        check_if(1 == 2)
    assert check_if(1 == 1, ForTestingErrorWithDoc) is None
    with pytest.raises(ForTestingErrorWithDoc, match="for testing purposes"):
        check_if(1 == 2, ForTestingErrorWithDoc)

def test_message_is_None_exception_with_docstring_asserts():
    with pytest.raises(AssertionError, match=""):
        assert_if(1 == 2, message=None)
    with pytest.raises(AssertionError, match="Error"):
        assert_if(1 == 2, message="Error")
    with pytest.raises(AssertionError):
        assert_if(1 == 2)
    assert assert_if(1 == 1, handle_with = ForTestingErrorWithDoc) is None
    with pytest.raises(ForTestingErrorWithDoc, match="for testing purposes"):
        assert_if(1 == 2, handle_with = ForTestingErrorWithDoc)

def test_message_is_None_exception_without_docstring():
    assert check_if(1 == 1, ForTestingErrorWithoutDoc) is None
    with pytest.raises(ForTestingErrorWithoutDoc, match="Error! Shout!"):
        check_if(1 == 2, ForTestingErrorWithoutDoc, message="Error! Shout!")
    with pytest.raises(ForTestingErrorWithoutDoc, match=""):
        check_if(1 == 2, ForTestingErrorWithoutDoc)

def test_message_is_None_exception_without_docstring():
    assert assert_if(1 == 1, handle_with = ForTestingErrorWithoutDoc) is None
    with pytest.raises(ForTestingErrorWithoutDoc, match="Error! Shout!"):
        assert_if(1 == 2,  handle_with = ForTestingErrorWithoutDoc, message="Error! Shout!")
    with pytest.raises(ForTestingErrorWithoutDoc, match=""):
        assert_if(1 == 2, handle_with = ForTestingErrorWithoutDoc)
